###############################################################################
# (c) Copyright 2021-2023 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Functor collections for the FunTuple framework.

A typical usage is the following:

.. code-block:: python

    coll_of_predefined_variables = Kinematics()

List of available collections (with contents):

{0}
"""

#import io, sys
from typing import List

from GaudiKernel import SystemOfUnits
from GaudiConf.LbExec import HltSourceID
import Functors as F
from PyConf.reading import get_odin, get_decreports, get_hlt1_selreports
from PyConf.application import default_raw_event, make_data_with_FetchDataFromFile
from PyConf.dataflow import DataHandle
from PyConf.components import Algorithm
from PyConf.Algorithms import HltTisTosAlg, Hlt1TrueSimEffAlg, Hlt2TrueSimEffAlg
import DaVinciMCTools  # for MCReconstructible, MCReconstructed and MCTruthAndBkgCat helper classes
from DecayTreeFitter import DecayTreeFitter
from .FunctorCollection import FunctorCollection

__all__ = ("EventInfo", "SelectionInfo", "HltTisTos", "Kinematics",
           "MCHierarchy", "MCKinematics", "MCVertexInfo", "MCPromptDecay",
           "MCReconstructible", "MCReconstructed", "TrackIsolation",
           "NeutralCaloInfo", "DecayTreeFitterResults", "ParticleID",
           "MCPrimaries")


def EventInfo():
    """
    Event-level collection of most-used functors for event information.
    By default the FunTuple framework stores the RUN_NUMBER
    and EVENT_NUMBER variables.
    When running over truth samples (MCParticles) i.e. xgen files,
    the RUN_NUMBER and EVENT_NUMBER are not yet available
    (see issue https://gitlab.cern.ch/lhcb/DaVinci/-/issues/90).

    Example:
        from FunTuple import FunTuple_Particles as Funtuple
        import FunTuple.functorcollections as FC
        from DaVinci import options

        variables = FC.EventInfo()
        # ...

        tuple = Funtuple(name="MyTuple",
                         fields=...,
                         variables=variables,
                         ...
                         )
    """
    odin = get_odin()

    event_info = {}
    event_info["BUNCHCROSSING_ID"] = F.BUNCHCROSSING_ID(odin)
    event_info["BUNCHCROSSING_TYPE"] = F.BUNCHCROSSING_TYPE(odin)
    event_info["ODINTCK"] = F.ODINTCK(odin)
    event_info["GPSTIME"] = F.GPSTIME(odin)

    return FunctorCollection(event_info)


def SelectionInfo(*, selection_type: HltSourceID,
                  trigger_lines: List[str]) -> FunctorCollection:
    """"
    Event-level collection of most-used functors for tupling trigger/Sprucing information.

    Args:
        selection_type (HltSourceID): Name of the selection type i.e. "Hlt1" or "Hlt2" or "Spruce".
            Used as branch name prefix when tupling and as source ID to get decision reports.
        trigger_lines (list(str)): List of line names for which the decision is requested.

    Example:
        from FunTuple import FunTuple_Particles as Funtuple
        from DaVinci import options
        import FunTuple.functorcollections as FC

        #List of lines you want decisions for
        selection_type = HltSourceiD.Hlt2
        hlt2_lines = ['Hlt2CharmD0ToKmPipLineDecision']

        variables = FC.SelectionInfo(selection_type, hlt2_lines)

        tuple = Funtuple(name="MyTuple",
                         fields=...,
                         variables=variables,
                         ...
                         )
    """
    selection_type = HltSourceID(selection_type)
    #get decreports
    dec_report = get_decreports(selection_type)
    #check that the code ends with decision
    add_suff_to_elem = lambda s: s + 'Decision' if not s.endswith('Decision') else s
    trigger_lines = list(map(add_suff_to_elem, trigger_lines))
    trigger_info = FunctorCollection({
        selection_type.name + "_TCK":
        F.TCK(dec_report)
    })
    trigger_info += FunctorCollection(
        {l: F.DECISION(dec_report, l)
         for l in trigger_lines})
    return trigger_info


def _HltMCTrueSimEff(
        *,
        selection_type: HltSourceID,
        trigger_lines: List[str],
        mcp_data: DataHandle,
        dec_reports: DataHandle,
        mc2IDLink: DataHandle,
        sel_reports: DataHandle = make_data_with_FetchDataFromFile(
            "dummy_sel_rep_location"),
        cand_locations: List[str] = ['dummy_line_name_Decision_loc']):
    """
    Candidate-level collection to store TrueSim efficiencies for HltEfficiencyChecker
    Configured in MooreAnalysis/HltEfficiencyChecker/python/HltEfficiencyChecker/config.py
      *Only* intended for usage in HltEfficienyChecker.
    These efficiencies are described in the HltEfficiencyChecker documentation:
      https://lhcbdoc.web.cern.ch/lhcbdoc/moore/master/tutorials/hltefficiencychecker.html
    Note that for HLT1, this is not configured by HltEfficiencyChecker due to MooreAnalysis#41.
      Use of this in HltEfficiencyChecker will currently crash.

    Args:
        selection_type (HltSourceID): Name of the selection type i.e. "Hlt1" or "Hlt2".
          Used as branch name prefix when tupling and as source ID to get decision reports.
        trigger_lines (list(str)): List of Hlt1/Hlt2 decision names for computing TrueSim efficiencies.
          Should be post-fixed with 'Decision'.
        mcp_data (DataHandle): TES location of MC particles e.g. output of
          `get_mc_particles("/Event/MC/Particles")
        dec_reports (DataHandle): TES location of DecReports.
        mc2IDLink (DataHandle): TES location of linker between MC particles and MC LHCbIDs of
          their tracks. Output of e.g. make_links_lhcbids_mcparticles_tracking_system()
        sel_reports (DataHandle, optional): TES location of Hlt1 SelReports. Only relevant for HLT1,
          so has a dummy default value.
        cand_locations (list(str), optional): List of TES locations to the reconstructed candidates
          of each line. Only relevant for HLT2, so has a dummy default value.

    Example:
        Only intended usage is in MooreAnalysis/HltEfficiencyChecker/python/HltEfficiencyChecker/config.py.
    """

    if selection_type == HltSourceID.Hlt1:
        true_sim_relns = Hlt1TrueSimEffAlg(
            name="Hlt1TrueSimEffAlgName_{hash}",
            InputMCParticles=mcp_data,
            DecReports=dec_reports,
            MC2IDLink=mc2IDLink,
            SelReports=sel_reports,
            TriggerLines=trigger_lines).P2TrueSimEffTable
    elif selection_type == HltSourceID.Hlt2:
        true_sim_relns = Hlt2TrueSimEffAlg(
            name="Hlt2TrueSimEffAlgName_{hash}",
            InputMCParticles=mcp_data,
            DecReports=dec_reports,
            MC2IDLink=mc2IDLink,
            TriggerLines=trigger_lines,
            CandidateLocations=cand_locations).P2TrueSimEffTable
    else:
        raise ValueError(
            f'selection_type "{selection_type}" is not supported in _HltMCTrueSimEff functor collection.'
        )

    return FunctorCollection({
        f"{line}TrueSim": F.VALUE_OR(False) @ F.VALUE_FROM_DICT(line)
        @ F.MAP_TO_RELATED(true_sim_relns)
        for line in trigger_lines
    })


def HltTisTos(*, selection_type: HltSourceID, trigger_lines: List[str],
              data: DataHandle) -> FunctorCollection:
    """
    Candidate-level collection to store TIS (Trigger Independent of Signal) or TOS (Trigger On Signal)
    information of candidates for specified Hlt1/Hlt2 lines.
    Note that the TIS/TOS information for Hlt2 lines is work-in-progress and currently only Hlt1 lines are supported.
    For TOB (Trigger On Both), the condition is (!TIS && !TOS && event_decision).
    To store "event decision" of a line use the `SelectionInfo` functor collection.

    Args:
        selection_type (HltSourceID): Name of the selection type i.e. "Hlt1" or "Hlt2". Currently only "Hlt1" is supported.
          Used as branch name prefix when tupling and as source ID to get decision reports.
        trigger_lines (list(str)): List of Hlt1/Hlt2 line names for
          which the decision is requested.
        data (DataHandle): TES location of particles from a Sprucing/Hlt2 line
          i.e. output of for example `get_particles("/Event/HLT2/<linename>/Particles")`.

    Example:
        from DaVinci import options
        import FunTuple.functorcollections as FC

        # get input data from a particular "<linename>" (can be Sprucing line or Hlt2 line)
        data = get_particles(f"/Event/HLT2/<linename>/Particles")

        #List of lines for which Tis/Tos information is requested
        trigger_lines = ["Hlt1TwoTrackMVADecision", "Hlt1TrackMVADecision"]
        variables = FC.HltTisTos(selection_type="Hlt1", trigger_lines=trigger_lines, data=data)

        #store event decision to check if a candidate is "TOB" wrt to a given line
        evt_variables = FC.SelectionInfo(selection_type="Hlt1", trigger_lines)

        tuple = Funtuple(name="MyTuple",
                         fields=...,
                         variables=variables,
                         event_variables=evt_variables,
                         ...
                         )
    """
    selection_type = HltSourceID(selection_type)
    if selection_type == HltSourceID.Hlt2:
        raise NotImplementedError(
            "The TIS/TOS information for Hlt2 lines is not yet implemented.")

    # Get the selection and decision reports
    hlt1_sel_report = get_hlt1_selreports()
    hlt1_dec = get_decreports(selection_type)

    TisTosRelations = HltTisTosAlg(
        name="HltTisTosAlgName_{hash}",
        InputParticles=data,
        SelReports=hlt1_sel_report,
        DecReports=hlt1_dec,
        TriggerLines=trigger_lines).P2TisTosTable

    hlt1tistos_info = FunctorCollection({
        f'{line}_TOS': F.IS_TOS(line, TisTosRelations)
        for line in trigger_lines
    })
    hlt1tistos_info += FunctorCollection({
        f'{line}_TIS': F.IS_TIS(line, TisTosRelations)
        for line in trigger_lines
    })
    return hlt1tistos_info


def Kinematics() -> FunctorCollection:
    """
    Candidate-level collection of most-used functors on kinematics.

    Example:
        from FunTuple import FunTuple_Particles as Funtuple
        import FunTuple.functorcollections as FC

        variables = FC.Kinematics()

        tuple = Funtuple(name="MyTuple",
                         fields=...,
                         variables=variables,
                         ...
                         )
    """
    kinematics = {
        "M": F.MASS,
        "P": F.P,
        "PT": F.PT,
        "PX": F.PX,
        "PY": F.PY,
        "PZ": F.PZ,
        "ENERGY": F.ENERGY,
    }
    return FunctorCollection(kinematics)


def MCKinematics(
        *, mctruth_alg: DaVinciMCTools.MCTruthAndBkgCat) -> FunctorCollection:
    """
    Candidate-level collection of most-used functors on kinematics on the related MC particle.

    Args:
        mctruth_alg (DaVinciMCTools.MCTruthAndBkgCat): An instance of the `MCTruthAndBkgCat` helper class
            that is used for retrieving MC truth information of a reconstructed particle.

    Example:
        from FunTuple import FunTuple_Particles as Funtuple
        import FunTuple.functorcollections as FC
        from DaVinciMCTools import MCTruthAndBkgCat

        #load particles onto TES location for a given event cycle
        input_data = get_particles("/Event/HLT2/<linename>/Particles")

        #define an algorithm that builds one-to-one relation table b/w Reco Particle -> Truth MC Particle.
        #Here "options" is the DV options object (see DaVinciExamples or DaVinciTutorials)
        mctruth_alg = MCTruthAndBkgCat(input_data)

        variables = FC.MCKinematics(mctruth_alg)

        tuple = Funtuple(name="MyTuple",
                         fields=...,
                         variables=variables,
        ...
        )
    """
    mckinematics = {
        "TRUEP": mctruth_alg(F.P),
        "TRUEPT": mctruth_alg(F.PT),
        "TRUEPX": mctruth_alg(F.PX),
        "TRUEPY": mctruth_alg(F.PY),
        "TRUEPZ": mctruth_alg(F.PZ),
        "TRUEENERGY": mctruth_alg(F.ENERGY),
    }
    return FunctorCollection(mckinematics)


def DecayTreeFitterResults(*,
                           DTF: DecayTreeFitter,
                           decay_origin: bool = True,
                           with_lifetime: bool = False,
                           with_kinematics: bool = True,
                           prefix: str = 'DTF') -> FunctorCollection:
    """
    Candidate-level collection of functors that access DecayTreeFitter fit results.

    Args:
        DTF (DecayTreeFitter.DecayTreeFitter): An instance of the `DecayTreeFitter` helper class
            that builds the necessary functors.
        decay_origin (bool): returns functorcollection for decay origin, defaults to True.
        with_lifetime (bool): The returned functors comprise lifetime and flight distance, defaults to False.
        with_kinematics (bool): The returned functors comprise kinematics, defaults to True.
        prefix (str): Additional prefix to output functors. Defaults to 'DTF'

    Example:
        from FunTuple import FunTuple_Particles as Funtuple
        import FunTuple.functorcollections as FC
        from DecayTreeFitter import DecayTreeFitter

        DTF = DecayTreeFitter(
            name = 'DTF',
            input_particles = get_particles(...),
            input_pvs = get_pvs()
        )

        variable_decay_origin = FC.DecayTreeFitter(DTF, decay_origin=True, with_lifetime=True, with_kinematics=True)
        variable_kinematics = FC.DecayTreeFitter(DTF, decay_origin=False, with_lifetime=False, with_kinematics=True)

        variables = {
            'B': variable_decay_origin,
            'phi_1020': variable_kinematics,
            'Kplus': variable_kinematics,
            'Kminus': variable_kinematics,
        }

        tuple = Funtuple(name="MyTuple",
                         fields=...,
                         variables=variables,
                         ...
                         )
    """
    suffix = f'[{prefix}_IDX]' if DTF.FitAllPVs else ''

    output_collection = FunctorCollection()

    if decay_origin:
        output_collection += FunctorCollection({
            f'{prefix}_CHI2{suffix}':
            DTF.CHI2,
            f'{prefix}_NDOF{suffix}':
            DTF.NDOF,
            f'{prefix}_CHI2DOF{suffix}':
            DTF.CHI2DOF,
            f'{prefix}_NITER{suffix}':
            DTF.NITER,
            f'{prefix}_MASS{suffix}':
            DTF.MASS,
            f'{prefix}_MASSERR{suffix}':
            DTF.MASSERR,
            f'{prefix}_P{suffix}':
            DTF.P,
            f'{prefix}_PERR{suffix}':
            DTF.PERR,
        })
        if DTF.HasPVConstraint:
            output_collection += FunctorCollection({
                f'{prefix}_PV_KEY{suffix}':
                F.VALUE_OR(-1) @ DTF(
                    F.VALUE_OR(-1) @ F.OBJECT_KEY @ F.ENDVERTEX),
                f'{prefix}_PV_X{suffix}':
                DTF(F.END_VX),
                f'{prefix}_PV_Y{suffix}':
                DTF(F.END_VY),
                f'{prefix}_PV_Z{suffix}':
                DTF(F.END_VZ),
            })

    if with_lifetime:
        output_collection += FunctorCollection({
            f'{prefix}_TAU{suffix}':
            DTF.TAU,
            f'{prefix}_TAUERR{suffix}':
            DTF.TAUERR,
            f'{prefix}_FD{suffix}':
            DTF.FD,
            f'{prefix}_FDERR{suffix}':
            DTF.FDERR,
        })

    if with_kinematics:
        output_collection += FunctorCollection({
            f'{prefix}_PX{suffix}':
            DTF(F.PX),
            f'{prefix}_PY{suffix}':
            DTF(F.PY),
            f'{prefix}_PZ{suffix}':
            DTF(F.PZ),
            f'{prefix}_PE{suffix}':
            DTF(F.ENERGY),
        })

    return output_collection


def MCPromptDecay(
        *,
        mctruth_alg: DaVinciMCTools.MCTruthAndBkgCat = None,
        max_lifetime: float = 1e-7 * SystemOfUnits.ns,
        store_longlived_particle_id: bool = True) -> FunctorCollection:
    """
    Candidate-level collection of functors to determine if a decay is prompt based on the true lifetime of its ancestors.
    If a decay does not have any long-lived ancestors, it is classified as prompt.

    This algorithm is based on the Runs 1 and 2
    MCTupleToolPrompt tool (https://gitlab.cern.ch/lhcb/Analysis/-/blob/run2-patches/Phys/DecayTreeTupleMC/src/MCTupleToolPrompt.cpp).

    If no argument is provided, i.e. mctruth_alg==None, this collection creates functors that apply to MCParticles;
    otherwise, the functors created are assumed to apply to Particles,
    which will find the associated MCParticle automatically using the input truth matching.

    Args:
        mctruth_alg (DaVinciMCTools.MCTruthAndBkgCat): An instance of the `MCTruthAndBkgCat` helper class
            that is used for retrieving MC truth information of a reconstructed particle. Defaults to None.
        max_lifetime (float, optional): maximum lifetime of short-lived particles (ns). Defaults to 1e-7 ns.
        store_longlived_particle_id (bool, optional): create functors to store the particle ID and container key of the first long-lived ancestor. Defaults to True.

    Example:
        from FunTuple import FunTuple_Particles as Funtuple
        import FunTuple.functorcollections as FC
        from DaVinciMCTools import MCTruthAndBkgCat

        #load particles onto TES location for a given event cycle
        input_data = get_particles("/Event/HLT2/<linename>/Particles")

        #define an algorithm that builds one-to-one relation table b/w Reco Particle -> Truth MC Particle.
        #Here "options" is the DV options object (see DaVinciExamples or DaVinciTutorials)
        mctruth_alg = MCTruthAndBkgCat(input_data)

        variables = FC.MCPromptDecay(mctruth_alg)

        tuple = Funtuple(name="MyTuple",
                         fields=...,
                         variables=variables,
                         inputs=input_data)
    """
    if mctruth_alg is None:
        MCTRUTH = lambda func: func
    else:
        MCTRUTH = mctruth_alg

    collection = FunctorCollection({
        "MC_ISPROMPT":
        F.VALUE_OR(False) @ MCTRUTH(F.MC_ISPROMPT(max_lifetime)),
    })
    if store_longlived_particle_id:
        collection += FunctorCollection({
            "MC_LONGLIVED_ID":
            F.VALUE_OR(0) @ MCTRUTH(
                F.VALUE_OR(0) @ F.PARTICLE_ID
                @ F.MC_FIRST_LONGLIVED_ANCESTOR(max_lifetime)),
            "MC_LONGLIVED_KEY":
            F.VALUE_OR(-1) @ MCTRUTH(
                F.VALUE_OR(-1) @ F.OBJECT_KEY
                @ F.MC_FIRST_LONGLIVED_ANCESTOR(max_lifetime)),
        })
    return collection


def MCHierarchy(*, mctruth_alg: DaVinciMCTools.MCTruthAndBkgCat = None
                ) -> FunctorCollection:
    """
    Candidate-level collection of functors yielding the hierarchy/relationships among MC particles.
    If no argument is provided, i.e. mctruth_alg==None, this collection creates functors that apply to MCParticles;
    otherwise, the functors created are assumed to apply to Particles,
    which will find the associated MCParticle automatically using the input truth matching.

    Args:
        mctruth_alg (DaVinciMCTools.MCTruthAndBkgCat): An instance of the `MCTruthAndBkgCat` helper class
            that is used for retrieving MC truth information of a reconstructed particle. Defaults to None.

    Example:
        from FunTuple import FunTuple_Particles as Funtuple
        import FunTuple.functorcollections as FC
        from DaVinciMCTools import MCTruthAndBkgCat

        #load particles onto TES location for a given event cycle
        input_data = get_particles("/Event/HLT2/<linename>/Particles")

        #define an algorithm that builds one-to-one relation table b/w Reco Particle -> Truth MC Particle.
        #Here "options" is the DaVinci options object (see DaVinciExamples or DaVinciTutorials package)
        mctruth_alg = MCTruthAndBkgCat(input_data)

        variables = FC.MCHierarchy(mctruth_alg)

        tuple = Funtuple(name="MyTuple",
                         fields=...,
                         variables=variables,
                         inputs=input_data)
    """
    if mctruth_alg is None:
        MCTRUTH = lambda func: func
    else:
        MCTRUTH = mctruth_alg

    MCMOTHER_ID = lambda n: F.VALUE_OR(0 ) @ MCTRUTH(F.MC_MOTHER(n, F.PARTICLE_ID))
    MCMOTHER_KEY= lambda n: F.VALUE_OR(-1) @ MCTRUTH(F.MC_MOTHER(n, F.OBJECT_KEY))
    MCHierarchy = {
        "TRUEID": F.VALUE_OR(0) @ MCTRUTH(F.PARTICLE_ID),
        "MC_MOTHER_ID": MCMOTHER_ID(1),
        "MC_MOTHER_KEY": MCMOTHER_KEY(1),
        "MC_GD_MOTHER_ID": MCMOTHER_ID(2),
        "MC_GD_MOTHER_KEY": MCMOTHER_KEY(2),
        "MC_GD_GD_MOTHER_ID": MCMOTHER_ID(3),
        "MC_GD_GD_MOTHER_KEY": MCMOTHER_KEY(3),
    }
    return FunctorCollection(MCHierarchy)


def MCVertexInfo(*, mctruth_alg: DaVinciMCTools.MCTruthAndBkgCat = None
                 ) -> FunctorCollection:
    """
    Candidate-level collection of functors on vertices of the related MC particle.

    Args:
        mctruth_alg (DaVinciMCTools.MCTruthAndBkgCat): An instance of the `MCTruthAndBkgCat` helper class
            that is used for retrieving monte-carlo truth information of a reconstructed particle. Defaults to None.

    Example:
        from FunTuple import FunTuple_Particles as Funtuple
        import FunTuple.functorcollections as FC
        from DaVinciMCTools import MCTruthAndBkgCat

        #load particles onto TES location for a given event cycle
        input_data = get_particles("/Event/HLT2/<linename>/Particles")

        #define an algorithm that builds one-to-one relation table b/w Reco Particle -> Truth MC Particle.
        #Here "options" is the DV options object (see DaVinciExamples or DaVinciTutorials)
        mctruth_alg = MCTruthAndBkgCat(input_data)

        variables = FC.MCVertexInfo(mctruth_alg)

        tuple = Funtuple(name="MyTuple",
                         fields=...,
                         variables=variables,
        ...
        )
    """
    if mctruth_alg is None:
        MCTRUTH = lambda func: func
    else:
        MCTRUTH = mctruth_alg

    MCVertexInfo = {
        "TRUEORIGINVERTEX_X": MCTRUTH(F.ORIGIN_VX),
        "TRUEORIGINVERTEX_Y": MCTRUTH(F.ORIGIN_VY),
        "TRUEORIGINVERTEX_Z": MCTRUTH(F.ORIGIN_VZ),
        "TRUEENDVERTEX_X": MCTRUTH(F.END_VX),
        "TRUEENDVERTEX_Y": MCTRUTH(F.END_VY),
        "TRUEENDVERTEX_Z": MCTRUTH(F.END_VZ),
    }
    return FunctorCollection(MCVertexInfo)


def MCPrimaries(*, mc_header: DataHandle) -> FunctorCollection:
    """
    Event-level collection of functors to store the properties (X,Y,Z,T) of MC primary vertices.

    Args:
        mc_header (DataHandle): The TES location of the `MCHeader` object,
            obtainded via the `PyConf.reading.get_mc_header` function.

    Example:
        from FunTuple import FunTuple_Particles as Funtuple
        import FunTuple.functorcollections as FC

        mc_header = get_mc_header()

        evt_variables = FC.MCPrimaries(mc_header)

        tuple = Funtuple(name="MyTuple",
                         fields=...,
                         variables=...,
                         event_variables=evt_variables,
        ...
        )
    """

    mcprimaries = {
        "MCPVX[MCPV_IDX]":
        F.MAP(F.VALUE_OR(F.NaN) @ F.X_COORDINATE @ F.POSITION)
        @ F.MC_ALLPVS(mc_header),
        "MCPVY[MCPV_IDX]":
        F.MAP(F.VALUE_OR(F.NaN) @ F.Y_COORDINATE @ F.POSITION)
        @ F.MC_ALLPVS(mc_header),
        "MCPVZ[MCPV_IDX]":
        F.MAP(F.VALUE_OR(F.NaN) @ F.Z_COORDINATE @ F.POSITION)
        @ F.MC_ALLPVS(mc_header),
        "MCPVT[MCPV_IDX]":
        F.MAP(F.VALUE_OR(-1) @ F.MC_VTX_TIME) @ F.MC_ALLPVS(mc_header),
        "MCPV_SIZE":
        F.SIZE_OF @ F.MC_ALLPVS(mc_header),
    }
    return FunctorCollection(mcprimaries)


def MCReconstructible(*,
                      mcreconstructible_alg: DaVinciMCTools.MCReconstructible,
                      extra_info: bool = False,
                      mctruth_alg: DaVinciMCTools.MCTruthAndBkgCat = None
                      ) -> FunctorCollection:
    """
    Candidate-level collection of functors to store reconstrutible information for the MC particle.

    A stable charged or neutral MCParticle can be reconstructible in one of the following categories:
        -1 = No MC classification possible
        0  = Outside detector acceptance
        1  = In acceptance but not reconstructible
        2  = Reconstructible as a Long charged track
        3  = Reconstructible as a Downstream charged track
        4  = Reconstructible as an Upstream charged track
        5  = Reconstructible as a T charged track
        6  = Reconstructible as a VELO charged track
        50 = Reconstructible as a Neutral particle

    Note that the support for stable neutral MCParticle is not present yet (Requires seperate functors).
    An MC decay to be reconstructible implies that all final state MCParticles are reconstructible.
    For more info on definition (Run1/2) see: https://lhcb-comp.web.cern.ch/analysis/davinci/v8/recrecdefinition.htm

    Args:
        mcreconstructible_alg (DaVinciMCTools.MCReconstructible): An instance of the `MCReconstructible` helper class
            that builds the necessary functors.
        extra_info (bool, optional): If False, returns a functor to ntuple reconstructible category for
            charged MCParticle. If True, returns extra information e.g. hasUT, hasVelo, etc. Default is False.
        mctruth_alg (DaVinciMCTools.MCTruthAndBkgCat): An instance of the `MCTruthAndBkgCat` helper class
            that is used for retrieving monte-carlo truth information of a reconstructed particle. Defaults to None.

    Example:
        from FunTuple import FunTuple_MCParticles as Funtuple
        import FunTuple.functorcollections as FC
        from DaVinciMCTools import MCReconstructible as MCRectible

        #load MCParticles onto TES location for a given event cycle
        input_data = get_particles("/Event/HLT2/MC/Particles")

        #define the helper class and get functor collection
        reconstructible_alg = MCRectible()
        variables = FC.MCReconstructible(reconstructible_alg)

        tuple = Funtuple(name="MyTuple",
                         fields=...,
                         variables=variables,
        ...
        )
    """
    if mctruth_alg is None:
        MCTRUTH = lambda func: func
    else:
        MCTRUTH = mctruth_alg

    #See above reconstructible categories or functor "MC_ChargeReconstructible" in "Simulation.h" in Rec.
    func_dict = {
        "MC_RECONSTRUCTIBLE": MCTRUTH(mcreconstructible_alg.Reconstructible)
    }
    if extra_info:
        func_dict.update({
            #Has sub-detector hits
            # https://gitlab.cern.ch/lhcb/LHCb/-/blob/master/Event/MCEvent/include/Event/MCTrackInfo.h
            "MC_HASVELO":
            MCTRUTH(mcreconstructible_alg.HasVelo),
            "MC_HASVELO_AND_T":
            MCTRUTH(mcreconstructible_alg.HasVeloAndT),
            "MC_HAST":
            MCTRUTH(mcreconstructible_alg.HasT),
            "MC_HAST1":
            MCTRUTH(mcreconstructible_alg.HasT1),
            "MC_HAST2":
            MCTRUTH(mcreconstructible_alg.HasT2),
            "MC_HAST3":
            MCTRUTH(mcreconstructible_alg.HasT3),
            "MC_HAST1X":
            MCTRUTH(mcreconstructible_alg.HasT1X),
            "MC_HAST2X":
            MCTRUTH(mcreconstructible_alg.HasT2X),
            "MC_HAST3X":
            MCTRUTH(mcreconstructible_alg.HasT3X),
            "MC_HAST1S":
            MCTRUTH(mcreconstructible_alg.HasT1S),
            "MC_HAST2S":
            MCTRUTH(mcreconstructible_alg.HasT2S),
            "MC_HAST3S":
            MCTRUTH(mcreconstructible_alg.HasT3S),
            "MC_HASUT":
            MCTRUTH(mcreconstructible_alg.HasUT),
            "MC_HASUT1":
            MCTRUTH(mcreconstructible_alg.HasUT1),
            "MC_HASUT2":
            MCTRUTH(mcreconstructible_alg.HasUT2),
            #Within sub-detector acceptance
            # https://gitlab.cern.ch/lhcb/LHCb/-/blob/master/Event/MCEvent/include/Event/MCTrackInfo.h
            "MC_ACCVELO":
            MCTRUTH(mcreconstructible_alg.AccVelo),
            "MC_ACCVELO_AND_T":
            MCTRUTH(mcreconstructible_alg.AccVeloAndT),
            "MC_ACCT":
            MCTRUTH(mcreconstructible_alg.AccT),
            "MC_ACCT1":
            MCTRUTH(mcreconstructible_alg.AccT1),
            "MC_ACCT2":
            MCTRUTH(mcreconstructible_alg.AccT2),
            "MC_ACCT3":
            MCTRUTH(mcreconstructible_alg.AccT3),
            "MC_ACCT1X":
            MCTRUTH(mcreconstructible_alg.AccT1X),
            "MC_ACCT2X":
            MCTRUTH(mcreconstructible_alg.AccT2X),
            "MC_ACCT3X":
            MCTRUTH(mcreconstructible_alg.AccT3X),
            "MC_ACCT1S":
            MCTRUTH(mcreconstructible_alg.AccT1S),
            "MC_ACCT2S":
            MCTRUTH(mcreconstructible_alg.AccT2S),
            "MC_ACCT3S":
            MCTRUTH(mcreconstructible_alg.AccT3S),
            "MC_ACCUT":
            MCTRUTH(mcreconstructible_alg.AccUT),
            "MC_ACCUT1":
            MCTRUTH(mcreconstructible_alg.AccUT1),
            "MC_ACCUT2":
            MCTRUTH(mcreconstructible_alg.AccUT2),
            ##These should be updated
            #"MC_HASVELOR": MCTRUTH(mcreconstructible_alg.HasVeloR),
            #"MC_HASVELOPhi": MCTRUTH(mcreconstructible_alg.HasVeloPhi),
            #"MC_ACCVELOR": MCTRUTH(mcreconstructible_alg.AccVeloR),
            #"MC_ACCVELOPhi": MCTRUTH(mcreconstructible_alg.AccVeloPhi),
        })
    return FunctorCollection(func_dict)


def MCReconstructed(*,
                    mcreconstructed_alg: DaVinciMCTools.MCReconstructed,
                    extra_info: bool = False) -> FunctorCollection:
    """
    Candidate-level collection of functors to store reconstrutible information for the MC particle.

    A stable charged or neutral MCParticle can be reconstructed in one of the following categories:
        -1 No MC classification possible (e.g. NO MC)
         0 Not reconstructed
         1 Reconstructed as a Long charged track
         2 Reconstructed as a Downstream charged track
         3 Reconstructed as an Upstream charged track
         4 Reconstructed as a T charged track
         5 Reconstructed as a VELO charged track
         6 Reconstructed as a 2D Velo track
         7 Reconstructed as a Muon track
        50 Reconstructed as a Neutral particle
        51 Reconstructed as a merged pi0
        52 Reconstructed as a type that doesn't fit other categories

    An MC decay to be reconstructed implies that all final state MCParticles are reconstructed.
    For more info on definition (Run1/2) see: https://lhcb-comp.web.cern.ch/analysis/davinci/v8/recrecdefinition.htm

    (since the method tuple->farray() only support vector<double> currently, all array will convert to array of floats)

    Args:
        mcreconstructed_alg (DaVinciMCTools.MCReconstructed): An instance of the `MCReconstructed` helper class
            that first maps MCParticle to reconstructed track and builds the necessary functors.
        extra_info (bool, optional): If False, returns a functor to ntuple reconstructed category for
            MCParticle. If True, returns extra information e.g. TrackType, hasUT, hasVelo, etc. Default is False.
            If True, returns extra information e.g. TrackChi2, GhostProb, etc. Default is False.

    Example:
        from FunTuple import FunTuple_MCParticles as Funtuple
        import FunTuple.functorcollections as FC
        from DaVinciMCTools import MCReconstructed as MCRected

        #load particles onto TES location for a given event cycle
        input_data = get_particles("/Event/HLT2/MC/Particles")

        #map MCParticle -> reconstructed track
        mcreconstructed_alg = MCRected(input_data)
        #get the variables
        variables = FC.MCReconstructed(mcreconstructed_alg)

        tuple = Funtuple(name="MyTuple",
                         fields=...,
                         variables=variables,
        ...
        )
    """
    suffix = ''
    if not mcreconstructed_alg.use_best_mcmatch:
        suffix = '[TRACK_INDX]'

    func_dict = {
        f"RECONSTRUCTED{suffix}": mcreconstructed_alg.Reconstructed,
    }
    if extra_info:
        func_dict.update({
            #For various track properties (e.g track type) see:
            # https://gitlab.cern.ch/lhcb/LHCb/-/blob/master/Event/TrackEvent/include/Event/TrackEnums.h
            f"TRACK_TYPE{suffix}":
            mcreconstructed_alg.TrackType,
            f"TRACK_FLAG{suffix}":
            mcreconstructed_alg.Flag,
            f"TRACK_HISTORY{suffix}":
            mcreconstructed_alg.History,
            #Common track properties
            f"TRACK_REFPOINT_X{suffix}":
            mcreconstructed_alg.ReferencePoint_X,
            f"TRACK_REFPOINT_Y{suffix}":
            mcreconstructed_alg.ReferencePoint_Y,
            f"TRACK_REFPOINT_Z{suffix}":
            mcreconstructed_alg.ReferencePoint_Z,
            f"TRACK_QOVERP{suffix}":
            mcreconstructed_alg.QOVERP,
            f"TRACK_CHI2{suffix}":
            mcreconstructed_alg.Chi2,
            f"TRACK_CHI2DOF{suffix}":
            mcreconstructed_alg.Chi2DOF,
            f"TRACK_NDOF{suffix}":
            mcreconstructed_alg.nDoF,
            f"TRACK_GHOSTPROB{suffix}":
            mcreconstructed_alg.GhostProbability,
            f"TRACK_HAST{suffix}":
            mcreconstructed_alg.HasT,
            f"TRACK_HASUT{suffix}":
            mcreconstructed_alg.HasUT,
            f"TRACK_HASVELO{suffix}":
            mcreconstructed_alg.HasVelo,
            f"TRACK_NHITS{suffix}":
            mcreconstructed_alg.nHits,
            f"TRACK_NVPHITS{suffix}":
            mcreconstructed_alg.nVPHits,
            f"TRACK_NUTHITS{suffix}":
            mcreconstructed_alg.nUTHits,
            f"TRACK_NFTHITS{suffix}":
            mcreconstructed_alg.nFTHits,
            #PID properties
            f"TRACK_PID_MU{suffix}":
            mcreconstructed_alg.PIDmu,
            f"TRACK_PID_PI{suffix}":
            mcreconstructed_alg.PIDpi,
            f"TRACK_PID_K{suffix}":
            mcreconstructed_alg.PIDk,
            f"TRACK_PID_P{suffix}":
            mcreconstructed_alg.PIDp,
            f"TRACK_PID_E{suffix}":
            mcreconstructed_alg.PIDe,
        })

    if not mcreconstructed_alg.use_best_mcmatch:
        func_dict[
            f'TRACK_MCASSOCWEIGHTS{suffix}'] = mcreconstructed_alg.MCAssoc_Weights
        #Note no suffix needed since we are returning simply a number representing
        # the number of reconstructed tracks associated to an MC particle
        # (in technical terms the size of object "range" from relations table)
        func_dict['N_RECO_TRACKS'] = mcreconstructed_alg.nRecoTracks

    return FunctorCollection(func_dict)


def TrackIsolation(*, iso_rel_table: Algorithm) -> FunctorCollection:
    """
    Candidate-level collection of functors on track isolation, using information related to the particles
    inside the cone around a given reference particle.

    Args:
        iso_rel_table (WeightedRelTableAlg): algorithm instance returning the required relations table
            that maps a reference particle (forming the head of the cone)
            to associated particles inside the cone itself.

    Example:
        from FunTuple import FunTuple_Particles as Funtuple
        from FunTuple.functorcollections import TrackIsolation
        from PyConf.Algorithms import WeightedRelTableAlg


        iso_rel_table = WeightedRelTableAlg(
           ReferenceParticles=...,
           InputCandidates=...,
           Cut=...,
           OutputLevel=INFO)

        variables = TrackIsolation(iso_rel_table)

        tuple = Funtuple(name="MyTuple",
           fields=...,
           variables=variables,
           ...)


    Todo:
        Exclude particles that are in the decay descriptor and therefore belong to the decay that is being looked for.
        At the moment WeightedRelTableAlg is making relations between the reference particle (head of the cone)
        and related particles (particles within the cone) without checking whenever the particles in the cone
        belong to the same decay of the head.
    """
    SUMCONE      = lambda func: F.SUMCONE(Functor=func, Relations=iso_rel_table.OutputRelations)
    ASYM         = lambda func: F.ASYM(Functor=func, Relations=iso_rel_table.OutputRelations)

    #get the set of relations from the relation table stored in the TES location
    RELS = F.RELATIONS.bind(
        F.TES(iso_rel_table.OutputRelations), F.FORWARDARGS)
    #get ETA and PHI from the reference particle
    REF_ETA = F.ETA @ F.FORWARDARG1
    REF_PHI = F.PHI @ F.FORWARDARG1
    #get ETA and PHI from the related particle (To side of relation table)
    REL_ETA = F.ETA @ F.TO @ F.FORWARDARG0
    REL_PHI = F.PHI @ F.TO @ F.FORWARDARG0
    #evaluate DELTA ETA between reference and related particles
    DELTA_ETA = REL_ETA - REF_ETA
    #evaluate DELTA PHI between reference and related particles
    DELTA_PHI = F.ADJUST_ANGLE @ (REL_PHI - REF_PHI)

    TrackIsolationVariables = {
        "HEAD_CMULT":
        F.VALUE_OR(0)
        @ F.MAP_INPUT_SIZE(Relations=iso_rel_table.OutputRelations),
        "HEAD_CP":
        SUMCONE(F.P),
        "HEAD_CPT":
        SUMCONE(F.PT),
        "HEAD_CPX":
        SUMCONE(F.PX),
        "HEAD_CPY":
        SUMCONE(F.PY),
        "HEAD_CPZ":
        SUMCONE(F.PZ),
        "HEAD_PASY":
        ASYM(F.P),
        "HEAD_PTASY":
        ASYM(F.PT),
        "HEAD_PXASY":
        ASYM(F.PX),
        "HEAD_PYASY":
        ASYM(F.PY),
        "HEAD_PZASY":
        ASYM(F.PZ),
        "HEAD_DETA":
        F.MAP(DELTA_ETA()).bind(RELS(), F.FORWARDARGS),
        "HEAD_DPHI":
        F.MAP(DELTA_PHI()).bind(RELS(), F.FORWARDARGS)
    }
    return FunctorCollection(TrackIsolationVariables)


def NeutralCaloInfo() -> FunctorCollection:
    """
    Candidate-level collection of functors on neutral calorimeter hypotheses information.
    Useful for example for neutral particle reconstruction efficiency or neutral PID studies.

    Output variables:
        'CaloTrackMatchChi2'               : 2D chi2 for Track-CaloCluster matching
        'CaloNeutralShowerShape'           : 2nd order moment of the cluster
        'CaloClusterMass'                  : CALO cluster Mass obtained from MergedPi0Alg
        'CaloNeutralEcalEnergy'            : The ECAL cluster energy associated to the neutral CaloHypo
        'CaloNeutral1To9EnergyRatio'       : Fraction of highest 1x1 cell to 3x3 cells forming the cluster for the Neutral CaloHypo
        'CaloNeutral4To9EnergyRatio'       : Fraction of highest 2x2 cell to 3x3 cells forming the cluster for the Neutral CaloHypo
        'CaloNeutralHcal2EcalEnergyRatio'  : The Hcal/Ecal energy ratio associated to the neutral calorimeter hypotheses
        'CaloNeutralID'                    : Bitwise information (32bits) of all Neutral seed CellIDs
        'CaloNumSaturatedCells'            : Number of saturated cells
    """

    _NeutralCaloInfo = {
        'CaloTrackMatchChi2': F.CLUSTERMATCH_CHI2,
        'CaloNeutralShowerShape': F.CALO_NEUTRAL_SHOWER_SHAPE,
        'CaloClusterMass': F.CALO_CLUSTER_MASS,
        'CaloNeutralEcalEnergy': F.CALO_NEUTRAL_ECAL_ENERGY,
        'CaloNeutral1To9EnergyRatio': F.CALO_NEUTRAL_1TO9_ENERGY_RATIO,
        'CaloNeutral4To9EnergyRatio': F.CALO_NEUTRAL_4TO9_ENERGY_RATIO,
        'CaloNeutralHcal2EcalEnergyRatio':
        F.CALO_NEUTRAL_HCAL2ECAL_ENERGY_RATIO,
        'CaloNeutralID': F.CALO_NEUTRAL_ID,
        'CaloNumSaturatedCells': F.CALO_NUM_SATURATED_CELLS,
    }
    return FunctorCollection(_NeutralCaloInfo)


def ParticleID(extra_info: bool = False) -> FunctorCollection:
    """
    Candidate-level collection of functors on global particle identification variables.

    extra_info (bool, optional) : If True, store DLLs in addition. Defaults to False.
    """
    pid_vars = {
        'PARTICLE_ID': F.PARTICLE_ID,
        'PROBNN_D': F.PROBNN_D,
        'PROBNN_E': F.PROBNN_E,
        'PROBNN_GHOST': F.PROBNN_GHOST,
        'PROBNN_K': F.PROBNN_K,
        'PROBNN_MU': F.PROBNN_MU,
        'PROBNN_P': F.PROBNN_P,
        'PROBNN_PI': F.PROBNN_PI
    }
    if extra_info:
        pid_vars.update({
            'PID_E': F.PID_E,
            'PID_K': F.PID_K,
            'PID_MU': F.PID_MU,
            'PID_P': F.PID_P,
            'PID_PI': F.PID_PI,
        })

    return FunctorCollection(pid_vars)


def __print_collection(coll_name):
    """Nicely print out the contents of a collection."""
    from PyConf.Algorithms import WeightedRelTableAlg
    from Functors.grammar import BoundFunctor
    from PyConf.reading import get_tes_root, upfront_decoder

    _truth_matching_excluded = [
        "_HltMCTrueSimEff", "MCReconstructible", "MCReconstructed",
        "MCPrimaries"
    ]

    print(f"{coll_name}:")
    coll = globals()[coll_name]

    #get functor collection
    if coll_name == 'EventInfo':
        with default_raw_event.bind(raw_event_format=0.5):
            f_dict = coll().functor_dict
    elif coll_name == 'SelectionInfo':
        with default_raw_event.bind(raw_event_format=0.5):
            f_dict = coll(
                selection_type=HltSourceID.Hlt2,
                trigger_lines=['dummy_line_name_Decision']).functor_dict
    elif 'MC' in coll_name and coll_name not in _truth_matching_excluded:
        with default_raw_event.bind(raw_event_format=0.5),\
             get_tes_root.bind(input_process="Hlt2"),\
              upfront_decoder.bind(source  ="Hlt2" ):
            dummy_dh = make_data_with_FetchDataFromFile("dummy3")
            dummy_alg = DaVinciMCTools.MCTruthAndBkgCat(
                dummy_dh, name='MCTruthAndBkgCat_MC')
        f_dict = coll(mctruth_alg=dummy_alg).functor_dict
    elif coll_name == 'TrackIsolation':
        dummy_dh1 = make_data_with_FetchDataFromFile("dummy4")
        dummy_dh2 = make_data_with_FetchDataFromFile("dummy5")
        dummy_alg = WeightedRelTableAlg(
            ReferenceParticles=dummy_dh1, InputCandidates=dummy_dh2)
        f_dict = coll(iso_rel_table=dummy_alg).functor_dict
    elif coll_name == 'MCReconstructible':
        dummy_dh = make_data_with_FetchDataFromFile("dummy6")
        dummy_dh.force_type('LHCb::MCProperty')
        r_tible = DaVinciMCTools.MCReconstructible(dummy_dh)
        f_dict = coll(
            mcreconstructible_alg=r_tible, extra_info=True).functor_dict
    elif coll_name == 'MCReconstructed':
        dummy_dh = make_data_with_FetchDataFromFile("dummy7")
        with get_tes_root.bind(input_process="Hlt2"):
            r_ted = DaVinciMCTools.MCReconstructed(dummy_dh)
            f_dict = coll(
                mcreconstructed_alg=r_ted, extra_info=True).functor_dict
    elif coll_name == 'HltTisTos':
        with default_raw_event.bind(raw_event_format=0.5, stream="dummy_strm"):
            dummy_dh = make_data_with_FetchDataFromFile("dummy8")
            f_dict = coll(
                selection_type=HltSourceID.Hlt1,
                trigger_lines=['dummy_line_name_Decision'],
                data=dummy_dh).functor_dict
    elif 'MCPrimaries' in coll_name:
        dummy_dh = make_data_with_FetchDataFromFile("dummy9")
        dummy_dh.force_type('LHCb::MCHeader')
        f_dict = coll(mc_header=dummy_dh).functor_dict
    elif coll_name == 'DecayTreeFitterResults':
        dummy_particles = make_data_with_FetchDataFromFile("dummy10")
        dummy_pvs = make_data_with_FetchDataFromFile("dummy11")
        dummy_pvs.force_type('LHCb::Event::PV::PrimaryVertexContainer')
        DTF = DecayTreeFitter(
            'DecayTreeFitter_{hash}',
            input_particles=dummy_particles,
            input_pvs=dummy_pvs)
        f_dict = coll(
            DTF=DTF,
            decay_origin=True,
            with_lifetime=True,
            with_kinematics=True).functor_dict
    elif coll_name == '_HltMCTrueSimEff':
        dummy_dh1 = make_data_with_FetchDataFromFile("dummy12")
        dummy_dh2 = make_data_with_FetchDataFromFile("dummy13")
        dummy_dh3 = make_data_with_FetchDataFromFile("dummy14")
        dummy_dh4 = make_data_with_FetchDataFromFile("dummy15")
        f_dict = coll(
            selection_type=HltSourceID.Hlt2,
            trigger_lines=['dummy_line_name_Decision'],
            mcp_data=dummy_dh1,
            dec_reports=dummy_dh2,
            mc2IDLink=dummy_dh3,
            sel_reports=dummy_dh4,
            cand_locations=['dummy_line_name_Decision_loc']).functor_dict
    else:
        f_dict = coll().functor_dict

    for k, v in f_dict.items():
        _v = v.code_repr() if isinstance(v, BoundFunctor) else v
        print(f"    {k:15}: {_v}")
    print()  # add an extra line for better viewing


# Temporarily direct output to a string to dynamically build the docstring below
#old_stdout = sys.stdout
#tmp_stdout = io.StringIO()
#sys.stdout = tmp_stdout

# for c in __all__:
#    __print_collection(c)  # all printout directed to variable tmp_stdout

# Build the docstring on the fly
#__doc__ = __doc__.format(tmp_stdout.getvalue())

# Do not forget to reset sys.stdout and clean-up!
#sys.stdout = old_stdout
#del old_stdout, tmp_stdout, c
